from cms.plugin_pool import plugin_pool

from django.utils.lru_cache import lru_cache


@lru_cache()
def get_plugin_fields(plugin_type):
    klass = get_plugin_class(plugin_type)
    opts = klass.model._meta.concrete_model._meta
    fields = opts.local_fields + opts.local_many_to_many
    return [field.name for field in fields]


@lru_cache()
def get_plugin_class(plugin_type):
    return plugin_pool.get_plugin(plugin_type)


@lru_cache()
def get_plugin_model(plugin_type):
    return get_plugin_class(plugin_type).model



from collections import namedtuple

from django.core.serializers import deserialize
from django.db import transaction
from django.utils.encoding import force_text
from django.utils.functional import cached_property

from cms.models import CMSPlugin

from .utils import get_plugin_model


BaseArchivedPlugin = namedtuple(
    'BaseArchivedPlugin',
    ['pk', 'creation_date', 'position', 'plugin_type', 'parent_id', 'data']
)


class ArchivedPlugin(BaseArchivedPlugin):

    @cached_property
    def model(self):
        return get_plugin_model(self.plugin_type)

    @cached_property
    def deserialized_instance(self):
        data = {
            'model': force_text(self.model._meta),
            'fields': self.data,
        }

        # TODO: Handle deserialization error
        return list(deserialize('python', [data]))[0]

    @transaction.atomic
    def restore(self, placeholder, language, parent=None):
        plugin_kwargs = {
            'pk': self.pk,
            'plugin_type': self.plugin_type,
            'placeholder': placeholder,
            'language': language,
            'parent': parent,
            'position': self.position,
        }

        if parent:
            plugin = parent.add_child(**plugin_kwargs)
        else:
            plugin = CMSPlugin.add_root(**plugin_kwargs)

        if self.plugin_type != 'CMSPlugin':
            _d_instance = self.deserialized_instance
            _d_instance.object._no_reorder = True
            plugin.set_base_attr(_d_instance.object)
            _d_instance.save()
        return plugin



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [
        ('cms', '0016_auto_20160608_1535'),
        ('sites', '0001_initial'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='PlaceholderAction',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('action', models.CharField(max_length=30, choices=[('add_plugin', 'Add plugin'), ('change_plugin', 'Change plugin'), ('delete_plugin', 'Delete plugin'), ('move_plugin', 'Move plugin'), ('move_out_plugin', 'Move out plugin'), ('move_in_plugin', 'Move in plugin'), ('move_plugin_out_to_clipboard', 'Move out to clipboard'), ('move_plugin_in_to_clipboard', 'Move in to clipboard'), ('add_plugins_from_placeholder', 'Add plugins from placeholder'), ('paste_plugin', 'Paste plugin'), ('paste_placeholder', 'Paste placeholder'), ('clear_placeholder', 'Clear placeholder')])),
                ('pre_action_data', models.TextField(blank=True)),
                ('post_action_data', models.TextField(blank=True)),
                ('language', models.CharField(max_length=5, choices=settings.LANGUAGES)),
                ('order', models.PositiveIntegerField(default=1)),
            ],
            options={
                'ordering': ['order'],
            },
        ),
        migrations.CreateModel(
            name='PlaceholderOperation',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('operation_type', models.CharField(max_length=30, choices=[('add_plugin', 'Add plugin'), ('change_plugin', 'Change plugin'), ('delete_plugin', 'Delete plugin'), ('move_plugin', 'Move plugin'), ('cut_plugin', 'Cut plugin'), ('paste_plugin', 'Paste plugin'), ('paste_placeholder', 'Paste placeholder'), ('add_plugins_from_placeholder', 'Add plugins from placeholder'), ('clear_placeholder', 'Clear placeholder')])),
                ('token', models.CharField(max_length=120, db_index=True)),
                ('origin', models.CharField(max_length=255, db_index=True)),
                ('language', models.CharField(max_length=5, choices=settings.LANGUAGES)),
                ('user_session_key', models.CharField(max_length=120, db_index=True)),
                ('date_created', models.DateTimeField(auto_now_add=True, verbose_name='date created', db_index=True)),
                ('is_applied', models.BooleanField(default=False)),
                ('is_archived', models.BooleanField(default=False)),
                ('site', models.ForeignKey(to='sites.Site')),
                ('user', models.ForeignKey(verbose_name='user', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'ordering': ['-date_created'],
                'get_latest_by': 'date_created',
            },
        ),
        migrations.AddField(
            model_name='placeholderaction',
            name='operation',
            field=models.ForeignKey(related_name='actions', to='djangocms_history.PlaceholderOperation'),
        ),
        migrations.AddField(
            model_name='placeholderaction',
            name='placeholder',
            field=models.ForeignKey(to='cms.Placeholder'),
        ),
        migrations.AlterUniqueTogether(
            name='placeholderaction',
            unique_together=set([('operation', 'order')]),
        ),
    ]



"""
Actions are internal to djangocms-history
A placeholder operation can have multiple actions.

Example:
    User moves plugin from placeholder A to placeholder B
    This creates a MOVE_PLUGIN placeholder operation and this operation
    contains two actions:
        MOVE_OUT_PLUGIN -> Move plugin out of placeholder A
        MOVE_IN_PLUGIN -> Move plugin into placeholder B
"""
from __future__ import unicode_literals


ADD_PLUGIN = 'add_plugin'
CHANGE_PLUGIN = 'change_plugin'
DELETE_PLUGIN = 'delete_plugin'
MOVE_PLUGIN = 'move_plugin'
MOVE_OUT_PLUGIN = 'move_out_plugin'
MOVE_IN_PLUGIN = 'move_in_plugin'
PASTE_PLUGIN = 'paste_plugin'
PASTE_PLACEHOLDER = 'paste_placeholder'
ADD_PLUGINS_FROM_PLACEHOLDER = 'add_plugins_from_placeholder'
CLEAR_PLACEHOLDER = 'clear_placeholder'

# This action is bound to the clipboard
# Its triggered when a plugin is moved from a placeholder
# into the clipboard
MOVE_PLUGIN_OUT_TO_CLIPBOARD = 'move_plugin_out_to_clipboard'

# This action is triggered when a plugin is moved from a placeholder
# into the clipboard. Its bound to the source placeholder.
MOVE_PLUGIN_IN_TO_CLIPBOARD = 'move_plugin_in_to_clipboard'



from functools import partial

from . import actions
from .helpers import get_bound_plugins, get_plugin_data


def _with_callback(func):
    """
    Runs an operation callback defined in the plugin class
    for the given operation handler.
    """
    def wrapped(*args, **kwargs):
        func(*args, **kwargs)

        if 'new_plugin' in kwargs:
            plugin = kwargs['new_plugin']
        else:
            plugin = kwargs['plugin']

        PluginClass = plugin.get_plugin_class()

        try:
            callbacks = PluginClass.operation_handler_callbacks
            callback = callbacks[func.__name__]
            callback(*args, **kwargs)
        except (AttributeError, KeyError):
            pass
    return wrapped


@_with_callback
def pre_add_plugin(operation, **kwargs):
    # Stores the ID of the parent plugin where the new plugin
    # will be created and the original order of the tree
    # where this plugin is being added
    plugin = kwargs['plugin']
    action_data = {
        'parent_id': plugin.parent_id,
        'order': kwargs['tree_order'],
    }

    operation.create_action(
        action=actions.ADD_PLUGIN,
        language=plugin.language,
        placeholder=kwargs['placeholder'],
        pre_data=action_data,
    )


@_with_callback
def post_add_plugin(operation, **kwargs):
    # Stores the ID of the parent plugin where the new plugin
    # will be created, the plugin data for the new created plugin,
    # and the new order of the tree where this plugin is being added
    plugin = kwargs['plugin']
    action_data = {
        'parent_id': plugin.parent_id,
        'plugins': [get_plugin_data(plugin=plugin)],
        'order': kwargs['tree_order'],
    }

    operation.set_post_action_data(action=actions.ADD_PLUGIN, data=action_data)


@_with_callback
def pre_change_plugin(operation, **kwargs):
    # Stores
    #   * the plugin data before any updates

    plugin = kwargs['old_plugin']
    action_data = {'plugins': [get_plugin_data(plugin=plugin)]}

    operation.create_action(
        action=actions.CHANGE_PLUGIN,
        language=plugin.language,
        placeholder=kwargs['placeholder'],
        pre_data=action_data,
    )


@_with_callback
def post_change_plugin(operation, **kwargs):
    # Stores
    #   * the plugin data with updates already applied

    plugin = kwargs['new_plugin']
    action_data = {
        'plugins': [get_plugin_data(plugin=plugin)]
    }

    operation.set_post_action_data(
        action=actions.CHANGE_PLUGIN,
        data=action_data,
    )


@_with_callback
def pre_delete_plugin(operation, **kwargs):
    # Stores
    #   * the id of the parent plugin for the plugin being deleted
    #   * plugin data for the plugin being deleted and all its descendants.
    #   * the tree order before the plugin got deleted

    get_data = get_plugin_data

    plugin = kwargs['plugin']
    descendants = plugin.get_descendants().order_by('path')
    plugin_data = [get_data(plugin=plugin)]
    plugin_data.extend(get_data(plugin) for plugin in get_bound_plugins(descendants))

    action_data = {
        'parent_id': plugin.parent_id,
        'plugins': plugin_data,
        'order': kwargs['tree_order'],
    }

    operation.create_action(
        action=actions.DELETE_PLUGIN,
        language=plugin.language,
        placeholder=kwargs['placeholder'],
        pre_data=action_data,
    )


@_with_callback
def post_delete_plugin(operation, **kwargs):
    # Stores
    #   * the id of the parent plugin for the deleted plugin
    #   * plugin meta data for the deleted plugin
    #   * the tree order after the plugin got deleted

    plugin = kwargs['plugin']
    plugin_data = [get_plugin_data(plugin=plugin, only_meta=True)]
    action_data = {
        'order': kwargs['tree_order'],
        'parent_id': plugin.parent_id,
        'plugins': plugin_data,
    }

    operation.set_post_action_data(action=actions.DELETE_PLUGIN, data=action_data)


def pre_move_plugin(operation, **kwargs):
    # Action 1 Stores
    #   * the tree order of the source placeholder before the plugin is moved
    #   * the id of the parent plugin for the plugin being moved
    #   * plugin meta data for the plugin being moved

    action_data = {
        'order': kwargs['source_order'],
        'parent_id': kwargs['source_parent_id'],
        'plugins': [get_plugin_data(plugin=kwargs['plugin'], only_meta=True)],
    }

    move_out = kwargs['source_placeholder'] != kwargs['target_placeholder']

    if move_out:
        action = actions.MOVE_OUT_PLUGIN
    else:
        action = actions.MOVE_PLUGIN

    operation.create_action(
        action=action,
        language=kwargs['source_language'],
        placeholder=kwargs['source_placeholder'],
        pre_data=action_data,
    )

    # Action 2 Stores
    #   * the tree order of the target placeholder before the plugin is moved
    #   * the id of the new parent plugin

    if move_out:
        action_data = {
            'order': kwargs['target_order'],
            'parent_id': kwargs['target_parent_id'],
        }

        operation.create_action(
            action=actions.MOVE_IN_PLUGIN,
            language=kwargs['target_language'],
            placeholder=kwargs['target_placeholder'],
            pre_data=action_data,
            order=2,
        )


def post_move_plugin(operation, **kwargs):
    # Action 1 Stores
    #   * the tree order of the target placeholder after the plugin was moved
    #   * the id of the new parent plugin
    #   * plugin meta data for the moved plugin

    action_data = {
        'order': kwargs['target_order'],
        'parent_id': kwargs['target_parent_id'],
        'plugins': [get_plugin_data(plugin=kwargs['plugin'], only_meta=True)],
    }

    move_in = kwargs['source_placeholder'] != kwargs['target_placeholder']

    if move_in:
        action = actions.MOVE_IN_PLUGIN
    else:
        action = actions.MOVE_PLUGIN

    operation.set_post_action_data(action=action, data=action_data)

    # Action 2 Stores
    #   * the tree order of the source placeholder after the plugin was moved
    #   * the id of the old parent plugin

    if move_in:
        action_data = {
            'order': kwargs['source_order'],
            'parent_id': kwargs['source_parent_id'],
        }
        operation.set_post_action_data(action=actions.MOVE_OUT_PLUGIN, data=action_data)


def pre_paste_plugin(operation, **kwargs):
    # Stores
    #   * the tree order of the target placeholder before the plugin is pasted

    action_data = {'order': kwargs['target_order']}

    operation.create_action(
        action=actions.PASTE_PLUGIN,
        language=kwargs['target_language'],
        placeholder=kwargs['target_placeholder'],
        pre_data=action_data,
    )


def post_paste_plugin(operation, **kwargs):
    # Stores
    #   * the tree order of the target placeholder after the plugin was pasted
    #   * the id of the new parent plugin
    #   * plugin data for the pasted plugin and all its descendants

    get_data = get_plugin_data

    plugin = kwargs['plugin']
    descendants = plugin.get_descendants().order_by('path')
    plugin_data = [get_data(plugin=plugin)]
    plugin_data.extend(get_data(plugin) for plugin in get_bound_plugins(descendants))
    action_data = {
        'order': kwargs['target_order'],
        'parent_id': kwargs['target_parent_id'],
        'plugins': plugin_data,
    }

    operation.set_post_action_data(action=actions.PASTE_PLUGIN, data=action_data)


def pre_paste_placeholder(operation, **kwargs):
    # Stores
    #   * the tree order of the target placeholder before the plugins are pasted

    action_data = {
        'order': kwargs['target_order'],
    }

    operation.create_action(
        action=actions.PASTE_PLACEHOLDER,
        language=kwargs['target_language'],
        placeholder=kwargs['target_placeholder'],
        pre_data=action_data,
    )


def post_paste_placeholder(operation, **kwargs):
    # Stores
    #   * the tree order of the target placeholder after the plugins were pasted
    #   * plugin data for the pasted plugins

    get_data = get_plugin_data
    plugins = get_bound_plugins(kwargs['plugins'])
    plugin_data = [get_data(plugin=plugin) for plugin in plugins]
    action_data = {'order': kwargs['target_order'], 'plugins': plugin_data}
    operation.set_post_action_data(action=actions.PASTE_PLACEHOLDER, data=action_data)


def pre_cut_plugin(operation, **kwargs):
    # Action 1 Stores
    #   * plugin meta data for the cut plugin

    get_data = get_plugin_data

    plugin = kwargs['plugin']
    plugin_data = [get_data(plugin=plugin, only_meta=True)]
    action_data = {'plugins': plugin_data}

    operation.create_action(
        action=actions.MOVE_PLUGIN_IN_TO_CLIPBOARD,
        language=kwargs['clipboard_language'],
        placeholder=kwargs['clipboard'],
        post_data=action_data,
        order=1,
    )

    # Action 2 Stores
    #   * the tree order of the source placeholder before the plugin is cut
    #   * the id of the parent plugin for the plugin being cut
    #   * plugin data for the plugin being cut and all its descendants

    descendants = plugin.get_descendants().order_by('path')

    plugins = [plugin]
    plugins.extend(get_bound_plugins(descendants))

    plugin_data = [get_data(plugin=plugin) for plugin in plugins]
    action_data = {
        'order': kwargs['source_order'],
        'parent_id': kwargs['source_parent_id'],
        'plugins': plugin_data,
    }

    operation.create_action(
        action=actions.MOVE_PLUGIN_OUT_TO_CLIPBOARD,
        language=kwargs['source_language'],
        placeholder=kwargs['source_placeholder'],
        pre_data=action_data,
        order=2,
    )


def post_cut_plugin(operation, **kwargs):
    # Stores
    #   * the tree order of the target placeholder after the plugin was cut
    #   * the id of the parent plugin for the cut plugin

    action_data = {
        'order': kwargs['source_order'],
        'parent_id': kwargs['source_parent_id'],
    }

    operation.set_post_action_data(
        action=actions.MOVE_PLUGIN_OUT_TO_CLIPBOARD,
        data=action_data,
    )


def pre_add_plugins_from_placeholder(operation, **kwargs):
    # Stores
    #   * plugin data for the pasted plugins

    action_data = {'order': kwargs['target_order']}

    operation.create_action(
        action=actions.ADD_PLUGINS_FROM_PLACEHOLDER,
        language=kwargs['target_language'],
        placeholder=kwargs['target_placeholder'],
        pre_data=action_data,
    )


def post_add_plugins_from_placeholder(operation, **kwargs):
    # Stores
    #   * plugin data for the new plugins
    #   * the tree order of the target placeholder after the new plugins are added

    get_data = get_plugin_data
    plugins = get_bound_plugins(kwargs['plugins'])
    plugin_data = [get_data(plugin=plugin) for plugin in plugins]
    action_data = {
        'plugins': plugin_data,
        'order': kwargs['target_order']
    }

    operation.set_post_action_data(
        action=actions.ADD_PLUGINS_FROM_PLACEHOLDER,
        data=action_data,
    )


def pre_clear_placeholder(operation, **kwargs):
    # Stores
    #   * plugin data for all the plugins being deleted

    get_data = get_plugin_data
    plugins = get_bound_plugins(kwargs['plugins'])
    plugin_data = [get_data(plugin=plugin) for plugin in plugins]
    action_data = {'plugins': plugin_data}

    operation.create_action(
        action=actions.CLEAR_PLACEHOLDER,
        language=operation.language,
        placeholder=kwargs['placeholder'],
        pre_data=action_data,
    )


def post_clear_placeholder(operation, **kwargs):
    # Stores
    #   * plugin meta data for all the deleted parent less  plugins

    get_data = partial(get_plugin_data, only_meta=True)
    root_plugins = [get_data(plugin=plugin) for plugin in kwargs['plugins']
                    if not plugin.parent_id]
    action_data = {'plugins': root_plugins}

    operation.set_post_action_data(action=actions.CLEAR_PLACEHOLDER, data=action_data)



from django.contrib import admin
from django.conf.urls import  url

from . import views
from .models import PlaceholderOperation


@admin.register(PlaceholderOperation)
class PlaceholderOperationAdmin(admin.ModelAdmin):

    def has_add_permission(self, request):
        return False

    def has_change_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False

    def get_urls(self):
        # This sucks but its our only way to register the internal
        # undo/redo urls without asking users to configure them
        urlpatterns = [
            url(r'^undo/$', views.undo, name='djangocms_history_undo'),
            url(r'^redo/$', views.redo, name='djangocms_history_redo'),
        ]
        return urlpatterns



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djangocms_history', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='placeholderaction',
            name='language',
            field=models.CharField(max_length=15, choices=settings.LANGUAGES),
        ),
        migrations.AlterField(
            model_name='placeholderoperation',
            name='language',
            field=models.CharField(max_length=15, choices=settings.LANGUAGES),
        ),
    ]



# -*- coding: utf-8 -*-
import json

from django.core.urlresolvers import reverse
from django.utils.translation import ugettext

from cms.constants import REFRESH_PAGE
from cms.toolbar_base import CMSToolbar
from cms.toolbar_pool import toolbar_pool
from cms.toolbar.items import BaseButton, ButtonList

from sekizai.helpers import get_varname

from .helpers import (
    get_active_operation,
    get_inactive_operation,
    get_operations_from_request,
)


class AjaxButton(BaseButton):
    template = 'djangocms_history/toolbar/ajax_button.html'

    def __init__(self, name, url, data, icon, active=False, disabled=False):
        self.name = name
        self.url = url
        self.active = active
        self.disabled = disabled
        self.data = data
        self.on_success = REFRESH_PAGE
        self.icon = icon

    def get_context(self):
        return {
            'name': self.name,
            'icon': self.icon,
            'active': self.active,
            'disabled': self.disabled,
            'data': json.dumps(self.data),
            'url': self.url,
            'on_success': self.on_success
        }


@toolbar_pool.register
class UndoRedoToolbar(CMSToolbar):
    undo_icon = (
        'data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAACAAAAAgCAYAAABzenr0AAABOklEQVR4A'
        'e3WNVZDQRhAYdyhimcBeMsG6CI1C8Dd9pK0rACXCm+QDoeaFnfnFlNNns4bnHvO10T/55Plsv/+i6ITM'
        'zjCtXCAOfFeBJ5axCb80h+n8YA3G49IIwSl3oRtBBDHKd5cukBSeQDhGC94U/SCXvUBjO2gF9UoFWoxi'
        'F2TIZI6BrhHF3JgVq4Y7sHgcIS8DnCEAJzUaDBESsch2IYfTuo1uDoiOs6BdTgp1+Cc6IBtqzYDLMBpg9J3p'
        '/Gp1UoDHOBTK5MGuMKnViFfjn/uEAx95UmofBnKFaAKbuszuBGFVf58Ag9IgLzfip2WjzHpYTSIXIWH0TmCcF'
        'UDXk0ex8OoRwXKUIchi8dxAkq1fMWCRK5JcUl2jgS0FMAInh0uSlPyMdc5SBdmsItLXGEfM+hwe6n99987pVTyg'
        'pWk5ykAAAAASUVORK5CYII='
    )
    redo_icon = (
        'data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAACAAAAAgCAYAAABzenr0AAABP0lEQVR42u'
        '3WP0vDQBiAcSXoYMWtpcYPoFjd3R01rn4Alx40Q9vv0q5+AxGsbiKOuoharM7OQv8MVqzng3Q6Ei5vLmQ'
        'xD/zGcC9ckruFoiJhPhSuMMBk7g0XUNhA5q2jiy9oiyk6xiBl3OMG4gKMoIU+cIAKnqABWSFm0CnN8A4NQ'
        'NBRzOJ9NFFDCSvYRB0PtqEkez6M2NsGPMS1CIVP1wG6EYvvI0llDFwG8CPe9oZg8UfXLVARe+4hSXdZvAOX'
        'xgNNJO3asvgtrL0aD20j18bGAKv/bgBzC2rILflLKM/9M3RP/iMKIW0Ly7Am/xXbCzDFedohqjGHUWjZDg8t'
        '4zA6wxLEBYg7jlvYQQlr2EUbz9CGH+xhXv4XkhM4dYhhyivZMTKpKriUfuMUFWSejzp6eMEYI/TRg/pbuKhI0'
        'C8BZPKBEtpCEgAAAABJRU5ErkJggg=='
    )
    icon_css = '<style>.cms-btn-disabled img {opacity: 0.2;}</style>'

    def populate(self):
        self.active_operation = self.get_active_operation()

        if self.toolbar.edit_mode:
            self.add_buttons()

    def get_operations(self):
        operations = get_operations_from_request(
            self.request,
            language=self.toolbar.language,
        )
        return operations

    def get_active_operation(self):
        operations = self.get_operations()
        return get_active_operation(operations)

    def get_inactive_operation(self):
        operations = self.get_operations()
        operation = get_inactive_operation(
            operations,
            active_operation=self.active_operation,
        )
        return operation

    def add_buttons(self):
        container = ButtonList(side=self.toolbar.RIGHT)
        container.buttons.append(self.get_undo_button())
        container.buttons.append(self.get_redo_button())
        self.toolbar.add_item(container)

    def _get_ajax_button(self, name, url, icon, disabled=True):
        data = {
            'language': self.toolbar.language,
            'cms_path': self.request.path,
            'csrfmiddlewaretoken': self.toolbar.csrf_token,
        }
        button = AjaxButton(
            name=name,
            url=url,
            data=data,
            icon=icon,
            active=False,
            disabled=disabled,
        )
        return button

    def get_undo_button(self):
        url = reverse('admin:djangocms_history_undo')
        disabled = not bool(self.active_operation)
        button = self._get_ajax_button(
            name=ugettext('Undo'),
            url=url,
            icon=self.undo_icon,
            disabled=disabled,
        )
        return button

    def get_redo_button(self):
        operation = self.get_inactive_operation()
        url = reverse('admin:djangocms_history_redo')
        disabled = not bool(operation)
        button = self._get_ajax_button(
            name=ugettext('Redo'),
            url=url,
            icon=self.redo_icon,
            disabled=disabled,
        )
        return button

    def render_addons(self, context):
        # TODO: This is a workaround to add our css to sekizai
        # Should be fixed on the cms core by exposing the context
        # to the toolbar populate methods.
        context[get_varname()]['css'].append(self.icon_css)
        return []



from django.core.exceptions import PermissionDenied
from django.http import HttpResponse, HttpResponseBadRequest
from django.views.generic import DetailView

from .forms import UndoRedoForm
from .helpers import (
    get_active_operation,
    get_inactive_operation,
    get_operations_from_request,
)
from .models import PlaceholderOperation


class UndoRedoView(DetailView):
    action = None
    model = PlaceholderOperation
    http_method_names = ['post']
    form_class = UndoRedoForm

    def post(self, request, *args, **kwargs):
        user = request.user

        if not (user.is_active and user.is_staff):
            raise PermissionDenied

        self.form = self.form_class(request.POST)

        if not self.form.is_valid():
            return HttpResponseBadRequest('No operation found')

        self.object = self.get_object()

        if not self.object:
            return HttpResponseBadRequest('No operation found')

        if self.action == 'undo':
            self.object.undo()
        else:
            self.object.redo()
        return HttpResponse(status=204)

    def get_object(self, queryset=None):
        if queryset is None:
            queryset = self.get_queryset()

        if self.action == 'undo':
            operation = get_active_operation(queryset)
        else:
            operation = get_inactive_operation(queryset)
        return operation

    def get_queryset(self):
        data = self.form.cleaned_data
        queryset = get_operations_from_request(
            self.request,
            path=data['cms_path'],
            language=data['language'],
        )
        return queryset


undo = UndoRedoView.as_view(action='undo')
redo = UndoRedoView.as_view(action='redo')



from cms.constants import PLUGIN_MOVE_ACTION
from cms.plugin_base import CMSPluginBase


# Patch the plugin action_options to force the cms
# to reload the page once a plugin has been moved.
# This is needed to update the undo/redo buttons.
# A better option is to update the buttons via js,
# this however will come in a later stage.
CMSPluginBase.action_options[PLUGIN_MOVE_ACTION] = {'requires_reload': True}



from __future__ import unicode_literals
import json
import functools
import operator

from django.conf import settings
from django.core.serializers.json import DjangoJSONEncoder
from django.contrib.auth.signals import user_logged_in
from django.contrib.sites.models import Site
from django.db import models
from django.db import transaction
from django.db.models import Q
from django.dispatch import receiver
from django.utils import six

from cms import operations
from cms.models import Placeholder
from cms.signals import (
    pre_obj_operation,
    pre_placeholder_operation,
    post_placeholder_operation,
)

from . import actions
from . import action_handlers
from . import operation_handlers
from .datastructures import ArchivedPlugin


dump_json = functools.partial(json.dumps, cls=DjangoJSONEncoder)
reduce = six.moves.reduce


# TODO: This will likely change into a class based pool integration
# to allow for custom operations and actions

_operation_handlers = {
    operations.ADD_PLUGIN: {
        'pre': operation_handlers.pre_add_plugin,
        'post': operation_handlers.post_add_plugin,
    },
    operations.CHANGE_PLUGIN: {
        'pre': operation_handlers.pre_change_plugin,
        'post': operation_handlers.post_change_plugin,
    },
    operations.DELETE_PLUGIN: {
        'pre': operation_handlers.pre_delete_plugin,
        'post': operation_handlers.post_delete_plugin,
    },
    operations.MOVE_PLUGIN: {
        'pre': operation_handlers.pre_move_plugin,
        'post': operation_handlers.post_move_plugin,
    },
    operations.CUT_PLUGIN: {
        'pre': operation_handlers.pre_cut_plugin,
        'post': operation_handlers.post_cut_plugin,
    },
    operations.PASTE_PLUGIN: {
        'pre': operation_handlers.pre_paste_plugin,
        'post': operation_handlers.post_paste_plugin,
    },
    operations.PASTE_PLACEHOLDER: {
        'pre': operation_handlers.pre_paste_placeholder,
        'post': operation_handlers.post_paste_placeholder,
    },
    operations.ADD_PLUGINS_FROM_PLACEHOLDER: {
        'pre': operation_handlers.pre_add_plugins_from_placeholder,
        'post': operation_handlers.post_add_plugins_from_placeholder,
    },
    operations.CLEAR_PLACEHOLDER: {
        'pre': operation_handlers.pre_clear_placeholder,
        'post': operation_handlers.post_clear_placeholder,
    },
}

_action_handlers = {
    actions.ADD_PLUGIN: {
        'undo': action_handlers.undo_add_plugin,
        'redo': action_handlers.redo_add_plugin,
    },
    actions.CHANGE_PLUGIN: {
        'undo': action_handlers.undo_change_plugin,
        'redo': action_handlers.redo_change_plugin,
    },
    actions.DELETE_PLUGIN: {
        'undo': action_handlers.undo_delete_plugin,
        'redo': action_handlers.redo_delete_plugin,
    },
    actions.MOVE_PLUGIN: {
        'undo': action_handlers.undo_move_plugin,
        'redo': action_handlers.redo_move_plugin,
    },
    actions.MOVE_OUT_PLUGIN: {
        'undo': action_handlers.undo_move_out_plugin,
        'redo': action_handlers.redo_move_out_plugin,
    },
    actions.MOVE_IN_PLUGIN: {
        'undo': action_handlers.undo_move_in_plugin,
        'redo': action_handlers.redo_move_in_plugin,
    },
    actions.MOVE_PLUGIN_OUT_TO_CLIPBOARD: {
        'undo': action_handlers.undo_move_plugin_out_to_clipboard,
        'redo': action_handlers.redo_move_plugin_out_to_clipboard,
    },
    actions.MOVE_PLUGIN_IN_TO_CLIPBOARD: {
        'undo': action_handlers.undo_move_plugin_in_to_clipboard,
        'redo': action_handlers.redo_move_plugin_in_to_clipboard,
    },
    actions.PASTE_PLUGIN: {
        'undo': action_handlers.undo_paste_plugin,
        'redo': action_handlers.redo_paste_plugin,
    },
    actions.PASTE_PLACEHOLDER: {
        'undo': action_handlers.undo_paste_placeholder,
        'redo': action_handlers.redo_paste_placeholder,
    },
    actions.ADD_PLUGINS_FROM_PLACEHOLDER: {
        'undo': action_handlers.undo_add_plugins_from_placeholder,
        'redo': action_handlers.redo_add_plugins_from_placeholder,
    },
    actions.CLEAR_PLACEHOLDER: {
        'undo': action_handlers.undo_clear_placeholder,
        'redo': action_handlers.redo_clear_placeholder,
    },
}


@receiver(user_logged_in, dispatch_uid='archive_old_operations')
def archive_old_operations(sender, request, user, **kwargs):
    """
    Archives all user operations that don't match the new user session
    """
    site = Site.objects.get_current(request)
    p_operations = (
        PlaceholderOperation
        .objects
        .filter(user=request.user, site=site)
        .exclude(user_session_key=request.session.session_key)
    )
    p_operations.update(is_archived=True)


@receiver(pre_obj_operation)
def pre_page_operation_handler(sender, **kwargs):
    operation_type = kwargs['operation']
    p_operations = PlaceholderOperation.objects.all()

    if operation_type == operations.PUBLISH_STATIC_PLACEHOLDER:
        # Fetch all operations which act on the published
        # static placeholder
        p_id = kwargs['obj'].draft_id
        p_operations = p_operations.filter(actions__placeholder=p_id)
    elif operation_type in operations.PAGE_TRANSLATION_OPERATIONS:
        # Fetch all operations which act on the translation only
        page = kwargs['obj']
        page.set_translations_cache()
        translation = kwargs['translation']
        page_urls = (page.get_absolute_url(lang) for lang in page._title_cache)
        p_operations = p_operations.filter(
            origin__in=page_urls,
            language=translation.language,
        )
    else:
        # Fetch all operations which act on a page including its children
        # for all languages of the page
        page = kwargs['obj']
        page.set_translations_cache()
        page_urls = (page.get_absolute_url(lang) for lang in page._title_cache)
        queries = [Q(origin__startswith=url) for url in page_urls]
        p_operations = p_operations.filter(reduce(operator.or_, queries))

    if kwargs['obj'].site_id:
        # Both cms.Page and cms.StaticPlaceholder have a site field
        # the site field on cms.StaticPlaceholder is optional though.
        p_operations = p_operations.filter(site=kwargs['obj'].site)

    # Archive all fetched operations
    p_operations.update(is_archived=True)


@receiver(pre_placeholder_operation)
def create_placeholder_operation(sender, **kwargs):
    """
    Creates the initial placeholder operation record
    """
    request = kwargs.pop('request')
    operation_type = kwargs.pop('operation')
    handler = _operation_handlers.get(operation_type, {}).get('pre')

    # Adding cms_history=0 to any of the operation endpoints
    # will prevent the recording of history for that one operation
    cms_history = request.GET.get('cms_history', True)
    cms_history = models.BooleanField().to_python(cms_history)

    if not handler or not cms_history:
        return

    # kwargs['language'] can be None if the user has not enabled
    # I18N or is not using i18n_patterns
    language = kwargs['language'] or settings.LANGUAGE_CODE

    operation = PlaceholderOperation.objects.create(
        operation_type=operation_type,
        token=kwargs['token'],
        origin=kwargs['origin'],
        language=language,
        user=request.user,
        user_session_key=request.session.session_key,
        site=Site.objects.get_current(request),
    )
    handler(operation, **kwargs)


@receiver(post_placeholder_operation)
def update_placeholder_operation(sender, **kwargs):
    """
    Updates the created placeholder operation record,
    based on the configured post operation handlers.
    """
    request = kwargs.pop('request')
    operation_type = kwargs.pop('operation')

    handler = _operation_handlers.get(operation_type, {}).get('post')

    # Adding cms_history=0 to any of the operation endpoints
    # will prevent the recording of history for that one operation
    cms_history = request.GET.get('cms_history', True)
    cms_history = models.BooleanField().to_python(cms_history)

    if not handler or not cms_history:
        return

    site = Site.objects.get_current(request)

    p_operations = PlaceholderOperation.objects.filter(
        site=site,
        user=request.user,
        user_session_key=request.session.session_key
    )

    operation = p_operations.get(token=kwargs['token'])

    # Run the placeholder operation handler
    handler(operation, **kwargs)

    # Mark the new operation as applied
    p_operations.filter(pk=operation.pk).update(is_applied=True)

    # Mark any operation from this user's session made on a separate path
    # or made on the current path but not applied as archived.
    p_operations.filter(
        ~Q(origin=kwargs['origin'])
        |Q(origin=kwargs['origin'], is_applied=False)
    ).update(is_archived=True)

    # Last, mark any operation made by another user on the current path
    # as archived.
    # TODO: This will need to change to allow for concurrent editing
    # Its actually better to get the affected placeholders
    # and archive any operations that contains those
    foreign_operations = (
        PlaceholderOperation
        .objects
        .filter(origin=kwargs['origin'], site=site)
        .exclude(user=request.user)
     )
    foreign_operations.update(is_archived=True)


class PlaceholderOperation(models.Model):

    OPERATION_TYPES = (
        (operations.ADD_PLUGIN, 'Add plugin'),
        (operations.CHANGE_PLUGIN, 'Change plugin'),
        (operations.DELETE_PLUGIN, 'Delete plugin'),
        (operations.MOVE_PLUGIN, 'Move plugin'),
        (operations.CUT_PLUGIN, 'Cut plugin'),
        (operations.PASTE_PLUGIN, 'Paste plugin'),
        (operations.PASTE_PLACEHOLDER, 'Paste placeholder'),
        (operations.ADD_PLUGINS_FROM_PLACEHOLDER, 'Add plugins from placeholder'),
        (operations.CLEAR_PLACEHOLDER, 'Clear placeholder'),
    )

    operation_type = models.CharField(max_length=30, choices=OPERATION_TYPES)
    token = models.CharField(max_length=120, db_index=True)
    origin = models.CharField(max_length=255, db_index=True)
    language = models.CharField(max_length=15, choices=settings.LANGUAGES)
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        verbose_name="user",
    )
    # Django uses 40 character session keys but other backends might use longer..
    user_session_key = models.CharField(max_length=120, db_index=True)
    date_created = models.DateTimeField(
        db_index=True,
        auto_now_add=True,
        verbose_name="date created",
    )
    is_applied = models.BooleanField(default=False)
    is_archived = models.BooleanField(default=False)
    site = models.ForeignKey(Site)

    class Meta:
        get_latest_by = "date_created"
        ordering = ['-date_created']

    def create_action(self, action, language, placeholder, **kwargs):
        pre_data = kwargs.pop('pre_data', '')

        if pre_data:
            pre_data = dump_json(pre_data)

        post_data = kwargs.pop('post_data', '')

        if post_data:
            post_data = dump_json(post_data)

        self.actions.create(
            action=action,
            pre_action_data=pre_data,
            post_action_data=post_data,
            language=language,
            placeholder=placeholder,
            **kwargs
        )

    def set_pre_action_data(self, action, data):
        self.actions.filter(action=action).update(pre_action_data=dump_json(data))

    def set_post_action_data(self, action, data):
        self.actions.filter(action=action).update(post_action_data=dump_json(data))

    @transaction.atomic
    def undo(self):
        for action in self.actions.order_by('order'):
            action.undo()
        self.is_applied = False
        self.save(update_fields=['is_applied'])

    @transaction.atomic
    def redo(self):
        for action in self.actions.order_by('-order'):
            action.redo()
        self.is_applied = True
        self.save(update_fields=['is_applied'])


class PlaceholderAction(models.Model):
    ACTION_CHOICES = (
        (actions.ADD_PLUGIN, 'Add plugin'),
        (actions.CHANGE_PLUGIN, 'Change plugin'),
        (actions.DELETE_PLUGIN, 'Delete plugin'),
        (actions.MOVE_PLUGIN, 'Move plugin'),
        (actions.MOVE_OUT_PLUGIN, 'Move out plugin'),
        (actions.MOVE_IN_PLUGIN, 'Move in plugin'),
        (actions.MOVE_PLUGIN_OUT_TO_CLIPBOARD, 'Move out to clipboard'),
        (actions.MOVE_PLUGIN_IN_TO_CLIPBOARD, 'Move in to clipboard'),
        (actions.ADD_PLUGINS_FROM_PLACEHOLDER, 'Add plugins from placeholder'),
        (actions.PASTE_PLUGIN, 'Paste plugin'),
        (actions.PASTE_PLACEHOLDER, 'Paste placeholder'),
        (actions.CLEAR_PLACEHOLDER, 'Clear placeholder'),
    )

    action = models.CharField(max_length=30, choices=ACTION_CHOICES)
    pre_action_data = models.TextField(blank=True)
    post_action_data = models.TextField(blank=True)
    placeholder = models.ForeignKey(to=Placeholder)
    language = models.CharField(max_length=15, choices=settings.LANGUAGES)
    operation = models.ForeignKey(to=PlaceholderOperation, related_name='actions')
    order = models.PositiveIntegerField(default=1)

    class Meta:
        ordering = ['order']
        unique_together = ('operation', 'order')

    def _object_version_data_hook(self, data):
        if data and 'pk' in data:
            return ArchivedPlugin(**data)
        return data

    def _get_parsed_data(self, raw_data):
        data = json.loads(
            raw_data,
            object_hook=self._object_version_data_hook,
        )
        return data

    def get_pre_action_data(self):
        return self._get_parsed_data(self.pre_action_data)

    def get_post_action_data(self):
        return self._get_parsed_data(self.post_action_data)

    @transaction.atomic
    def undo(self):
        _action_handlers[self.action]['undo'](self)

    @transaction.atomic
    def redo(self):
        _action_handlers[self.action]['redo'](self)



from django import forms
from django.conf import settings


class UndoRedoForm(forms.Form):
    language = forms.ChoiceField(
        choices=settings.LANGUAGES,
        required=True,
    )
    cms_path = forms.CharField(required=True)



from collections import defaultdict
from datetime import timedelta

from django.contrib.sites.models import Site
from django.core import serializers
from django.core.exceptions import ObjectDoesNotExist
from django.db.models import signals
from django.utils import timezone

from cms.models import CMSPlugin
from cms.signals import (
    pre_delete_plugins,
    pre_save_plugins,
    post_delete_plugins
)
from cms.utils import get_language_from_request

from .utils import get_plugin_fields, get_plugin_model


def delete_plugins(placeholder, plugin_ids, nested=True):
    # With plugins, we can't do queryset.delete()
    # because this would trigger a bunch of internal
    # cms signals.
    # Instead, delete each plugin individually and turn off
    # position reordering using the _no_reorder trick.
    plugins = (
        placeholder
        .cmsplugin_set
        .filter(pk__in=plugin_ids)
        .order_by('-depth')
        .select_related()
    )

    bound_plugins = get_bound_plugins(plugins)

    for plugin in bound_plugins:
        plugin._no_reorder = True

        if hasattr(plugin, 'cmsplugin_ptr'):
            plugin.cmsplugin_ptr._no_reorder = True

        # When the nested option is False
        # avoid queries by preventing the cms from
        # recalculating the child counter of this plugin's
        # parent (for which there's none).
        plugin.delete(no_mp=not nested)


def get_bound_plugins(plugins):
    plugin_types_map = defaultdict(list)
    plugin_lookup = {}

    # make a map of plugin types, needed later for downcasting
    for plugin in plugins:
        plugin_types_map[plugin.plugin_type].append(plugin.pk)

    for plugin_type, pks in plugin_types_map.items():
        plugin_model = get_plugin_model(plugin_type)
        plugin_queryset = plugin_model.objects.filter(pk__in=pks)

        # put them in a map so we can replace the base CMSPlugins with their
        # downcasted versions
        for instance in plugin_queryset.iterator():
            plugin_lookup[instance.pk] = instance

    for plugin in plugins:
        yield plugin_lookup.get(plugin.pk, plugin)


def get_plugin_data(plugin, only_meta=False):
    if only_meta:
        custom_data = None
    else:
        plugin_fields = get_plugin_fields(plugin.plugin_type)
        _plugin_data = serializers.serialize('python', (plugin,), fields=plugin_fields)[0]
        custom_data = _plugin_data['fields']

    plugin_data = {
        'pk': plugin.pk,
        'creation_date': plugin.creation_date,
        'position': plugin.position,
        'plugin_type': plugin.plugin_type,
        'parent_id': plugin.parent_id,
        'data': custom_data,
    }
    return plugin_data


def get_active_operation(operations):
    operations = operations.filter(is_applied=True)

    try:
        operation = operations.latest()
    except ObjectDoesNotExist:
        operation = None
    return operation


def get_inactive_operation(operations, active_operation=None):
    active_operation = active_operation or get_active_operation(operations)

    if active_operation:
        date_created = active_operation.date_created
        operations = operations.filter(date_created__gt=date_created)

    try:
        operation = operations.filter(is_applied=False).earliest()
    except ObjectDoesNotExist:
        operation = None
    return operation


def get_operations_from_request(request, path=None, language=None):
    from .models import PlaceholderOperation

    if not language:
        language = get_language_from_request(language)

    origin = path or request.path

    # This is controversial :/
    # By design, we don't let undo/redo span longer than a day.
    # To be decided if/how this should be configurable.
    date = timezone.now() - timedelta(days=1)

    site = Site.objects.get_current(request)

    queryset = PlaceholderOperation.objects.filter(
        site=site,
        origin=origin,
        language=language,
        user=request.user,
        user_session_key=request.session.session_key,
        date_created__gt=date,
        is_archived=False,
    )
    return queryset


def disable_cms_plugin_signals(func):
    # The wrapped function NEEDS to set _no_reorder on any bound plugin instance
    # otherwise this does nothing because it only disconnects signals
    # for the cms.CMSPlugin class, not its subclasses
    plugin_signals = (
        (signals.pre_delete, pre_delete_plugins, 'cms_pre_delete_plugin', CMSPlugin),
        (signals.pre_save, pre_save_plugins, 'cms_pre_save_plugin', CMSPlugin),
        (signals.post_delete, post_delete_plugins, 'cms_post_delete_plugin', CMSPlugin),
    )

    def wrapper(*args, **kwargs):
        for signal, handler, dispatch_id, model_class in plugin_signals:
            signal.disconnect(
                handler,
                sender=model_class,
                dispatch_uid=dispatch_id
            )
            signal.disconnect(handler, sender=model_class)

        func(*args, **kwargs)

        for signal, handler, dispatch_id, model_class in plugin_signals:
            signal.connect(
                handler,
                sender=model_class,
                dispatch_uid=dispatch_id
            )

    return wrapper



from cms.models import CMSPlugin
from cms.utils.plugins import reorder_plugins

from .helpers import delete_plugins, disable_cms_plugin_signals


@disable_cms_plugin_signals
def _delete_plugins(action, plugin_ids, nested=True):
    delete_plugins(
        placeholder=action.placeholder,
        plugin_ids=plugin_ids,
        nested=nested,
    )
    action.placeholder.mark_as_dirty(action.language, clear_cache=False)


def _reorder_plugins(action, parent_id=None, order=None):
    if not order:
        return

    reorder_plugins(
        action.placeholder,
        parent_id=parent_id,
        language=action.language,
        order=order,
    )


@disable_cms_plugin_signals
def _restore_archived_plugins(action, data, root_plugin_id=None):
    plugins_by_id = {}

    if root_plugin_id:
        plugins_by_id[root_plugin_id] = CMSPlugin.objects.get(pk=root_plugin_id)

    for archived_plugin in data['plugins']:
        if archived_plugin.parent_id:
            parent = plugins_by_id[archived_plugin.parent_id]
        else:
            parent = None

        plugin = archived_plugin.restore(
            placeholder=action.placeholder,
            language=action.language,
            parent=parent,
        )
        plugins_by_id[plugin.pk] = plugin

    action.placeholder.mark_as_dirty(action.language, clear_cache=False)


@disable_cms_plugin_signals
def _restore_archived_plugins_tree(action, data, root_plugin_id=None):
    plugin_ids = [plugin.pk for plugin in data['plugins']]
    plugins_by_id = CMSPlugin.objects.in_bulk(plugin_ids)
    plugin_data = {'language': action.language, 'placeholder': action.placeholder}

    if root_plugin_id:
        root = CMSPlugin.objects.get(pk=root_plugin_id)
    else:
        root = None

    for _plugin in data['plugins']:
        plugin = plugins_by_id[_plugin.pk]

        if root:
            plugin = plugin.update(refresh=True, parent=root, **plugin_data)
            plugin = plugin.move(root, pos='last-child')
        else:
            target = CMSPlugin.get_last_root_node()
            plugin = plugin.update(refresh=True, parent=None, **plugin_data)
            plugin = plugin.move(target, pos='right')

        # Update all children to match the parent's
        # language and placeholder
        plugin.get_descendants().update(**plugin_data)

    action.placeholder.mark_as_dirty(action.language, clear_cache=False)


def undo_add_plugin(action):
    post_data = action.get_post_action_data()
    parent_id = post_data['parent_id']
    tree_order = action.get_pre_action_data()['order']

    # Only delete plugins who are direct children (or parent-less) of the
    # target parent.
    # This allows for cascade delete of the children of these plugins.
    plugin_ids = [plugin.pk for plugin in post_data['plugins']
                  if plugin.parent_id == parent_id]
    _delete_plugins(action, plugin_ids=plugin_ids, nested=bool(parent_id))
    _reorder_plugins(action, parent_id=parent_id, order=tree_order)


def redo_add_plugin(action):
    post_data = action.get_post_action_data()
    parent_id = post_data['parent_id']
    _restore_archived_plugins(
        action,
        data=post_data,
        root_plugin_id=parent_id,
    )
    _reorder_plugins(action, parent_id=parent_id, order=post_data['order'])


def undo_change_plugin(action):
    archived_plugins = action.get_pre_action_data()['plugins']

    for plugin in archived_plugins:
        if plugin.data:
            plugin.model.objects.filter(pk=plugin.pk).update(**plugin.data)


def redo_change_plugin(action):
    archived_plugins = action.get_post_action_data()['plugins']

    for plugin in archived_plugins:
        if plugin.data:
            plugin.model.objects.filter(pk=plugin.pk).update(**plugin.data)


def undo_delete_plugin(action):
    pre_data = action.get_pre_action_data()
    parent_id = pre_data['parent_id']
    _restore_archived_plugins(
        action,
        data=pre_data,
        root_plugin_id=parent_id,
    )
    _reorder_plugins(action, parent_id=parent_id, order=pre_data['order'])


def redo_delete_plugin(action):
    post_data = action.get_post_action_data()
    parent_id = post_data['parent_id']
    plugin_ids = [plugin.pk for plugin in post_data['plugins']]
    _delete_plugins(action, plugin_ids=plugin_ids, nested=bool(parent_id))
    _reorder_plugins(action, parent_id=parent_id, order=post_data['order'])


def undo_move_plugin(action):
    pre_data = action.get_pre_action_data()
    parent_id = pre_data['parent_id']
    _restore_archived_plugins_tree(
        action,
        data=pre_data,
        root_plugin_id=parent_id,
    )
    _reorder_plugins(
        action,
        parent_id=parent_id,
        order=pre_data['order'],
    )


def redo_move_plugin(action):
    post_data = action.get_post_action_data()
    parent_id = post_data['parent_id']
    _restore_archived_plugins_tree(
        action,
        data=post_data,
        root_plugin_id=parent_id,
    )
    _reorder_plugins(
        action,
        parent_id=parent_id,
        order=post_data['order'],
    )


def undo_move_in_plugin(action):
    pre_data = action.get_pre_action_data()
    _reorder_plugins(
        action,
        parent_id=pre_data['parent_id'],
        order=pre_data['order'],
    )


def redo_move_in_plugin(action):
    post_data = action.get_post_action_data()
    parent_id = post_data['parent_id']
    _restore_archived_plugins_tree(
        action,
        data=post_data,
        root_plugin_id=parent_id,
    )
    _reorder_plugins(
        action,
        parent_id=parent_id,
        order=post_data['order'],
    )


def undo_move_out_plugin(action):
    pre_data = action.get_pre_action_data()
    parent_id = pre_data['parent_id']
    _restore_archived_plugins_tree(
        action,
        data=pre_data,
        root_plugin_id=parent_id,
    )
    _reorder_plugins(
        action,
        parent_id=parent_id,
        order=pre_data['order'],
    )


def redo_move_out_plugin(action):
    post_data = action.get_post_action_data()
    _reorder_plugins(
        action,
        parent_id=post_data['parent_id'],
        order=post_data['order'],
    )


def undo_move_plugin_in_to_clipboard(action):
    # clear the clipboard
    action.placeholder.clear()


def redo_move_plugin_in_to_clipboard(action):
    post_data = action.get_post_action_data()

    # clear the clipboard
    action.placeholder.clear()

    # Add the plugin back to the clipboard
    # by restoring the data which points it to the clipboard
    # placeholder.
    _restore_archived_plugins_tree(action, data=post_data)


def undo_move_plugin_out_to_clipboard(action):
    pre_data = action.get_pre_action_data()
    parent_id = pre_data['parent_id']

    # Plugin was moved to the clipboard
    # Add it back to the source placeholder
    _restore_archived_plugins(
        action,
        data=pre_data,
        root_plugin_id=parent_id,
    )
    _reorder_plugins(
        action,
        parent_id=parent_id,
        order=pre_data['order'],
    )


def redo_move_plugin_out_to_clipboard(action):
    post_data = action.get_post_action_data()
    parent_id = post_data['parent_id']
    _reorder_plugins(
        action,
        parent_id=parent_id,
        order=post_data['order'],
    )


def undo_paste_plugin(action):
    tree_order = action.get_pre_action_data()['order']
    post_data = action.get_post_action_data()
    parent_id = post_data['parent_id']
    plugin_ids = (plugin.pk for plugin in post_data['plugins'])
    _delete_plugins(action, plugin_ids=plugin_ids, nested=bool(parent_id))
    _reorder_plugins(action, parent_id=parent_id, order=tree_order)


def redo_paste_plugin(action):
    post_data = action.get_post_action_data()
    parent_id = post_data['parent_id']
    _restore_archived_plugins(
        action,
        data=post_data,
        root_plugin_id=parent_id,
    )
    _reorder_plugins(
        action,
        parent_id=parent_id,
        order=post_data['order'],
    )


def undo_paste_placeholder(action):
    tree_order = action.get_pre_action_data()['order']
    post_data = action.get_post_action_data()
    plugin_ids = (plugin.pk for plugin in post_data['plugins'])
    _delete_plugins(action, plugin_ids=plugin_ids, nested=False)
    _reorder_plugins(action, parent_id=None, order=tree_order)


def redo_paste_placeholder(action):
    post_data = action.get_post_action_data()
    _restore_archived_plugins(action, data=post_data)
    _reorder_plugins(action, parent_id=None, order=post_data['order'])


def undo_add_plugins_from_placeholder(action):
    tree_order = action.get_pre_action_data()['order']
    post_data = action.get_post_action_data()
    plugin_ids = (plugin.pk for plugin in post_data['plugins'])
    _delete_plugins(action, plugin_ids=plugin_ids, nested=False)
    _reorder_plugins(action, parent_id=None, order=tree_order)


def redo_add_plugins_from_placeholder(action):
    post_data = action.get_post_action_data()
    _restore_archived_plugins(action, data=post_data)
    _reorder_plugins(action, parent_id=None, order=post_data['order'])


def undo_clear_placeholder(action):
    pre_data = action.get_pre_action_data()
    _restore_archived_plugins(action, data=pre_data)


def redo_clear_placeholder(action):
    action.placeholder.clear(action.language)



from setuptools import find_packages, setup

import djangocms_history


INSTALL_REQUIREMENTS = [
    'Django>=1.8,<1.11',
    'django-cms>=3.4.2',
    'django-sekizai>=0.7',
]


setup(
    name='djangocms-history',
    packages=find_packages(),
    include_package_data=True,
    version=djangocms_history.__version__,
    description=djangocms_history.__doc__,
    long_description=open('README.rst').read(),
    classifiers=[
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Topic :: Software Development'
    ],
    install_requires=INSTALL_REQUIREMENTS,
    author='Divio AG',
    author_email='info@divio.ch',
    url='http://github.com/divio/djangocms-history',
    license='BSD',
)
